#include <math.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_cdf.h>
#include <gsl/gsl_sf_gamma.h>
#include "gsl_distfit_eval.h"


/* Should maybe base everything more haevily on histograms?? */
/* W is akward */
/* Only approximate - center values approximated by left edge values. */
void gsl_distfit_init_moments(size_t n, double *x, double *px, double *_m1, double *_m2) {
  size_t i;
  double m1 = 0;
  double m2 = 0;
  double lower,upper,w;
  
  w = 0;
  lower = x[0];
  for (i=0; i < (n-1); i++) {
    upper = x[i+1];
    w     = upper - lower;
    
    m1 += px[i]*x[i]*w;
    m2 += gsl_pow_2(x[i])*px[i]*w;

    lower = upper;
  }

  *_m1 = m1 + px[n-1]*x[n-1]*w;
  *_m2 = m2 + px[n-1]*gsl_pow_2(x[n-1])*w;
}




/*p(x) dx = {1 \over \sqrt{2 \pi \sigma^2}} \exp (-(x - mu)^2 / 2\sigma^2) dx*/
double eval_gaussian(double x, const gsl_vector *p) {
  double mu     = gsl_vector_get(p,0);
  double sigma2 = gsl_pow_2(gsl_vector_get(p,1));

  return 1.0 / (sqrt(2*M_PI*sigma2)) * exp( - gsl_pow_2(x - mu)*0.5/sigma2);
}


double eval_gaussian_dsigma(double x, const gsl_vector *p) {
  double mu     = gsl_vector_get(p,0);
  double sigma2 = gsl_pow_2(gsl_vector_get(p,1));
  
  return exp(-gsl_pow_2(x - mu)*0.5/sigma2)/(sigma2*sqrt(2*M_PI))*(gsl_pow_2(x-mu)/sigma2 - 1);
}


double eval_gaussian_dmu(double x, const gsl_vector *p) {
  double mu     = gsl_vector_get(p,0);
  double sigma2 = gsl_pow_2(gsl_vector_get(p,1));

  return exp(-gsl_pow_2(x - mu)*0.5/sigma2)*sqrt(2)*(x - mu)/(sigma2*sqrt(sigma2*M_PI));
}


double eval_gaussian_Pinv(double x, const gsl_vector *p) {
  double mu    = gsl_vector_get(p,0);
  double sigma = gsl_vector_get(p,1);
  return gsl_cdf_gaussian_Pinv(x-mu,sigma);
}


void init_gaussian(size_t n, double *x, double *px, gsl_vector *p) {
  double m1, m2;
  gsl_distfit_init_moments(n,x,px,&m1,&m2);
  gsl_vector_set(p,0,m1);
  gsl_vector_set(p,1,sqrt(m2 - m1*m1));
}




/* <Plain functions>  */

/*p(x) dx = (x / sigma^2) exp(-x^2/(2 sigma^2)) dx*/

double eval_rayleigh(double x, const gsl_vector *p) {
  double inv_sigma_sqr = 1 / gsl_pow_2(gsl_vector_get(p,0));
  return x * inv_sigma_sqr * exp(-0.5*x*x*inv_sigma_sqr);
}

double eval_rayleigh_dsigma(double x, const gsl_vector *p) {
  double sigma = gsl_vector_get(p,0);
  return ((gsl_pow_2(x / sigma) - 2) * x / gsl_pow_3(sigma) * exp(-0.5*gsl_pow_2(x/sigma)));
}


double eval_rayleigh_Pinv(double x, const gsl_vector *p) {
  double sigma = gsl_vector_get(p,0);
  return gsl_cdf_rayleigh_Pinv(x,sigma);
}


/*
  
  m1 = sigma * sqrt(pi/2);
  m2 = 2*sigma^2;
  
*/

void init_rayleigh(size_t n, double *x, double *px, gsl_vector *p) {
  double m1, m2;
  gsl_distfit_init_moments(n,x,px,&m1,&m2);
  gsl_vector_set(p,0,0.5*(m1 * sqrt(2 / M_PI) + sqrt(m2 / 2)));
}

/* The lognormal distribution has the form 

   p(x) dx = 1/(x * sqrt(2 pi sigma^2)) exp(-(ln(x) - zeta)^2/2 sigma^2) dx

   for x > 0. Lognormal random numbers are the exponentials of
   gaussian random numbers */


/*
  m1 = exp(zeta + sigma*sigma/2);
  m2 = exp(2*(zeta + sigma*sigma));
*/
void init_lognormal(size_t n, double *x, double *px, gsl_vector *p) {
  double m1,m2;
  double zeta,sigma;
  
  gsl_distfit_init_moments(n,x,px,&m1,&m2);
  zeta  = 2 * log(m1) - 0.5*log(m2);
  sigma = sqrt(log(m2) - 2*log(m1));
  gsl_vector_set(p,0,zeta);
  gsl_vector_set(p,1,sigma);
}



double eval_lognormal(double x, const gsl_vector *p) {
  double zeta  = gsl_vector_get(p,0);
  double sigma = gsl_vector_get(p,1);
  
  return sqrt(2/(M_PI*sigma*sigma))*exp(-(pow(log(x) - zeta,2)/(2*pow(sigma,2))))/(2 * x);
}


double eval_lognormal_dsigma(double x, const gsl_vector *p) {
  double zeta  = gsl_vector_get(p,0);
  double sigma = gsl_vector_get(p,1);

  double e = exp(-(pow(log(x) - zeta,2)/(2*pow(sigma,2)))) * sqrt(2 / (M_PI * pow(sigma,2)));
  return e*(pow(log(x) - zeta,2) / (2*x*pow(sigma,3)) - 1/(2*x*sigma));
}


double eval_lognormal_dzeta(double x, const gsl_vector *p) {
  double zeta  = gsl_vector_get(p,0);
  double sigma = gsl_vector_get(p,1);
  
  return eval_lognormal(x,p)*(log(x) - zeta) / pow(sigma,2);
}


double eval_lognormal_Pinv(double x, const gsl_vector *p) {
  double zeta  = gsl_vector_get(p,0);
  double sigma = gsl_vector_get(p,1);
  
  return gsl_cdf_lognormal_Pinv(x,zeta,sigma);
}



/* The Weibull distribution has the form,

  p(x) dx = (b/a) (x/a)^(b-1) exp(-(x/a)^b) dx

*/
double eval_weibull(double x, const gsl_vector *p) {
  double a = gsl_vector_get(p,0);
  double b = gsl_vector_get(p,1);

  return (b / a) * pow(x / a,b-1) * exp(-pow(x/a,b));
}

double eval_weibull_da(double x, const gsl_vector *p) {
  double a = gsl_vector_get(p,0);
  double b = gsl_vector_get(p,1);

  double exp_x_a_b = exp(-pow(x/a,b));
  double x_a_b_1   = pow(x/a,b-1);
  double inv_a2    = 1/pow(a,2);

  return exp_x_a_b * inv_a2 * x_a_b_1 * b*b*(pow(x/a,b) - 1);
}


double eval_weibull_db(double x, const gsl_vector *p) {
  double a = gsl_vector_get(p,0);
  double b = gsl_vector_get(p,1);

  double exp_x_a_b = exp(-pow(x/a,b));
  double x_a_b_1   = pow(x/a,b-1);
  double inv_a     = 1/a;
  
  return exp_x_a_b * x_a_b_1 * inv_a*(1 + b*log(x/a)*(1 - pow(x/a,b)));
}


double eval_weibull_Pinv(double x, const gsl_vector *p) {
  double a = gsl_vector_get(p,0);
  double b = gsl_vector_get(p,1);
  return gsl_cdf_weibull_Pinv(x,a,b);
}


/*
  m1 = beta   * sf_gamma(1 + 1/alpha);
  m2 = beta^2 * sf_gamma(1 + 2/alpha);
*/
void init_weibull(size_t n, double *x, double *px, gsl_vector *p) {
  double m1,m2;
  double alpha,beta;

  gsl_distfit_init_moments(n,x,px,&m1,&m2);

  /* 
     For no apparent reason - I did not want to set up a nonlinear solver with Gamma functions.
  */
  alpha = sqrt(m2 - m1*m1);
  beta  = 2;
  
  gsl_vector_set(p,0,alpha);
  gsl_vector_set(p,1,beta);
  fprintf(stderr,"Warning: initialization of the weibull fit is completely broken\n");
}


double eval_exponential(double x, const gsl_vector *p) {
  double mu = gsl_vector_get(p,0);
  return 1.0 / mu * exp(-x/mu);
}


double eval_exponential_dmu(double x, const gsl_vector *p) {
  double mu = gsl_vector_get(p,0);
  return exp(-x/mu)*(x - 1)/gsl_pow_2(mu) ;
}


double eval_exponential_Pinv(double x, const gsl_vector *p) {
  double mu = gsl_vector_get(p,0);
  return gsl_cdf_exponential_Pinv(x,mu);
}

void init_exponential(size_t n, double *x, double *px, gsl_vector *p) {
  double m1, m2;
  gsl_distfit_init_moments(n,x,px,&m1,&m2);
  gsl_vector_set(p,0,m1);
}

/*
  p(x) dx = {1 \over 2 \Gamma(\nu/2) } (x/2)^{\nu/2 - 1} \exp(-x/2) dx
*/

double eval_dgamma(double x) {
  const double h0 = 0.05;
  double h;

  if (x > h0)
    h = h0;
  else
    h = 0.5*x;

  return (gsl_sf_gamma(x + h) - gsl_sf_gamma(x - h))*0.5/h;
}


double eval_chisq(double x, const gsl_vector *p) {
  double nu = gsl_vector_get(p,0);
  return  0.5 / gsl_sf_gamma(nu/2) * pow(x/2 , nu/2 - 1)*exp(-x/2);
}


double eval_chisq_dnu(double x, const gsl_vector *p) {
  double nu = gsl_vector_get(p,0);
  return exp(-x/2)*pow(2,-nu/2)*pow(x,nu/2 - 1)/gsl_pow_2(gsl_sf_gamma(nu/2))*(log(x)/2*gsl_sf_gamma(nu/2) - 0.5*eval_dgamma(nu/2));
}


void init_chisq(size_t n, double *x, double *px, gsl_vector *p) {
  double m1, m2;
  gsl_distfit_init_moments(n,x,px,&m1,&m2);
  printf("Kjorer init_chisq m1 = %g \n",m1);
  gsl_vector_set(p,0,m1);
}

double eval_chisq_Pinv(double x, const gsl_vector *p) {
  double nu = gsl_vector_get(p,0);
  return gsl_cdf_chisq_Pinv(x,nu);
}



/*
p(x) dx = {\Gamma((\nu + 1)/2) \over \sqrt{\pi \nu} \Gamma(\nu/2)}
(1 + x^2/\nu)^{-(\nu + 1)/2} dx*/

double eval_tdist(double x, const gsl_vector *p) {
  double nu = gsl_vector_get(p,0);
  return gsl_sf_gamma(0.5*(nu + 1))/(sqrt(M_PI * nu) * gsl_sf_gamma(nu*0.5))*pow(1 + gsl_pow_2(x)/nu,-0.5*(nu + 1));
}


double eval_tdist_dnu(double x, const gsl_vector *p) {
  double nu              = gsl_vector_get(p,0);
  double gamma_nu_half   = gsl_sf_gamma(nu/2);
  double gamma_nup1_half = gsl_sf_gamma((nu+1)/2);
  double sqrt_nu_pi      = sqrt(nu*M_PI);
  double power_arg       = 1 + gsl_pow_2(x)/nu;

  double gamma_factor    = gamma_nup1_half/(sqrt_nu_pi * gamma_nu_half);
  double power_factor    = pow(power_arg , -0.5*(1 + nu));

  double dgamma_factor   = (eval_dgamma(0.5*(nu + 1))*sqrt_nu_pi*gamma_nu_half - gamma_nup1_half*(eval_dgamma(nu*0.5)*sqrt_nu_pi + 0.5*sqrt(M_PI/nu)*gamma_nu_half)) / 
                           (nu*M_PI*gsl_pow_2(gamma_nu_half));
  double dpower_factor   = power_factor*((nu + 1)*0.5*gsl_pow_2(x/nu)/power_arg - 0.5*log(power_arg));

    
  return gamma_factor*dpower_factor + dgamma_factor*power_factor;
}


void init_tdist(size_t n, double *x, double *px, gsl_vector *p) {
  const double nu_max = 100;
  double m1, m2;
  double nu;
  gsl_distfit_init_moments(n,x,px,&m1,&m2);
  
  nu = 2*m2/(m2 - 1);
  if (nu > nu_max) 
    nu = nu_max;

  gsl_vector_set(p, 0 , nu);
}


double eval_tdist_Pinv(double x, const gsl_vector *p) { 
  double nu = gsl_vector_get(p,0);
  return gsl_cdf_tdist_Pinv(x,nu);
}


/*
  p(x) dx = {\Gamma(a+b) \over \Gamma(a) \Gamma(b)} x^{a-1} (1-x)^{b-1} dx
*/

double eval_beta(double x, const gsl_vector *p) {
  double a = gsl_vector_get(p,0);
  double b = gsl_vector_get(p,1);
  return gsl_sf_gamma(a+b) / (gsl_sf_gamma(a) * gsl_sf_gamma(b)) * pow(x,a-1)* pow(1-x,b-1);
}


double eval_beta_da(double x, const gsl_vector *p) {
  double a = gsl_vector_get(p,0);
  double b = gsl_vector_get(p,1);
  double gamma_a      = gsl_sf_gamma(a);
  double gamma_b      = gsl_sf_gamma(b);
  double gamma_ab     = gsl_sf_gamma(a+b);
  
  double power_xa_1   = pow(x,a-1);
  double power_1_xb_1 = pow(1-x,b-1);
  
  double power_factor = power_xa_1 * power_1_xb_1;
  double gamma_factor = gamma_ab / (gamma_a * gamma_b);
  
  return (eval_dgamma(a+b)*(gamma_a * gamma_b) - eval_dgamma(a) * gamma_b * gamma_ab)/(gsl_pow_2(gamma_a * gamma_b)) * power_factor + 
         gamma_factor * (power_factor * log(x));
}


double eval_beta_db(double x, const gsl_vector *p) {
  double a = gsl_vector_get(p,0);
  double b = gsl_vector_get(p,1);
  double gamma_a      = gsl_sf_gamma(a);
  double gamma_b      = gsl_sf_gamma(b);
  double gamma_ab     = gsl_sf_gamma(a+b);
  
  double power_xa_1   = pow(x,a-1);
  double power_1_xb_1 = pow(1-x,b-1);
  
  double power_factor = power_xa_1 * power_1_xb_1;
  double gamma_factor = gamma_ab / (gamma_a * gamma_b);
  
  return (eval_dgamma(a+b)*(gamma_a * gamma_b) - eval_dgamma(b) * gamma_a * gamma_ab)/(gsl_pow_2(gamma_a * gamma_b)) * power_factor + 
          gamma_factor * (power_factor * log(1-x));
}


void init_beta(size_t n, double *x, double *px, gsl_vector *p) {
  double m1, m2;
  double a,b;
  gsl_distfit_init_moments(n,x,px,&m1,&m2);
  a = 1;
  b = 1;
  gsl_vector_set(p,0,a);
  gsl_vector_set(p,1,b);
  fprintf(stderr," Warning: init of beta distribution is *not* implemented - using dummy values a = b = 1 \n");
  /* Solve these equations:
     m1 = a / (a + b);
     m2 = 1 / (a+b)^2 * (a*b/(a+b+1) + a*a)
  */

}

double eval_beta_Pinv(double x, const gsl_vector *p) {
  /*double a = gsl_vector_get(p,0);
    double b = gsl_vector_get(p,1);*/
  fprintf(stderr," Sorry Pinv is not implemented for the beta distribution - aborting\n");
  exit(1);
  return 0;
  /*return gsl_cdf_beta_Pinv(x,a,b);*/
}













